program Apply_Wavelet

  use omp_lib

  use sep
  use helix
  use verbosity_mod
  use adaptive_init
  use findadaptive_mod

  implicit none

  real,    dimension(:), allocatable   :: data,out
  real,    dimension(:), allocatable   :: wavelet
  type (filter)                        :: match

  integer, dimension(:), allocatable   :: nfilt

  integer, dimension(:), pointer :: n                  ! data-size, dimension(ndim)
  real,    dimension(:), pointer :: d, o

  integer, dimension(:), pointer :: idx                ! index counter (ndim)

  integer :: n1w,n12,i,ndim,stat,idim,ipanels,npanels,sided   ! junk variables
  character(len=1024)    :: label

  call sep_init(SOURCE)
 
  write(0,*) 'INFO:'
  write(0,*) 'INFO:'
  write(0,*) 'INFO:'
  write(0,*) 'INFO: -----------------------------------------------------'
  write(0,*) 'INFO: Convolving source with time-domain                   '
  write(0,*) 'INFO: estimated wavelet from Wavelet_Estimation            '
  write(0,*) 'INFO: with option average=0                                '
  write(0,*) 'INFO: -----------------------------------------------------'
  write(0,*) 'INFO:'

  ndim=sep_dimension()
  allocate(nfilt(1))

  if (ndim.le.2) then
     allocate(n(3), o(3), d(3))
     do i=1,ndim
        call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
     end do
     n(ndim+1:)=1
     o(ndim+1:)=0.
     d(ndim+1:)=1
  else
     allocate(n(ndim), o(ndim), d(ndim))
     do i=1,ndim
        call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
     end do
  end if

  if (exist_file('wavelet')) then
     call from_aux('wavelet','n1',n1w)
     call from_aux('wavelet','n2',npanels)
  else
     call erexit('ERROR: need input_wavelet file')
  end if
  if (npanels.ne.product(n(3:))) call erexit('ERROR: npanels in wavelet is not equal to product(n(3:)) in input data')


  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:       Dimensions Observed Data      '
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  do i=1,ndim
     write(0,*) 'INFO: - - - - - - - '
     write(0,*) 'INFO: n(',i,')=',n(i)
     write(0,*) 'INFO: o(',i,')=',o(i)
     write(0,*) 'INFO: d(',i,')=',d(i)
  end do
  write(0,*) 'INFO: Size of filter in # samples =',n1w

  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  n12=n(1)*n(2)
  nfilt=n1w

  call from_param('sided',sided,1) ! defaults to one-sided filter
  call create_match_filter_sided(nfilt,match,1,n,sided)
     
  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO: Starting processing of ',npanels
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  allocate(data(n12))
  allocate(out(n12))

  PANELS:do i=1,npanels

     call sreed('in',data,4*n12)
     call sreed('wavelet',match%flt,4*n1w)
     call applymatch(data,out,match)
     call srite('out',out,4*n12)
     
  end do PANELS
  
call sep_close()

end program APPLY_WAVELET
