!Antoine Guitton
!Log decon program in the Fourier domain 
!04/03/2012

program Sparse_decon

  use sep
  use sfft_mod
  use hysdstep_mod
  use quantile_calc_mod

  implicit none

  real,       dimension(:,:),   allocatable :: in      ! input data
  real,       dimension(:,:),   allocatable :: gt      ! gain function
  real,       dimension(:,:),   allocatable :: qt,dqt  ! gained residual and delta residual
  real,       dimension(:,:),   allocatable :: rt      ! residual time domain

  real,       dimension(:),     allocatable :: tmp1

  complex,    dimension(:,:),   allocatable :: inw     ! input Fourier domain
  complex,    dimension(:,:),   allocatable :: rw      ! residual Fourier domain
  complex,    dimension(:),     allocatable :: uw, duw, uw1 ! filter Fourier domain

  double complex, dimension(:), allocatable :: euw, euw1     ! filter and delta filter Fourier domain

  complex,    dimension(:),     allocatable :: tmp,tmp2 
  complex,    dimension(:),     allocatable :: drw     ! delta residual Fourier domain

  real,       dimension(:),     allocatable :: ut, ut1, dut ! filter time domain


  integer,    dimension(:),     allocatable :: n
  real,       dimension(:),     allocatable :: o, d
  integer(8), dimension(:),     allocatable :: plan1
  logical,    parameter                     :: FORWARD=.FALSE.,BACKWARD=.TRUE.

  character(len=1024) :: label

  integer :: i,j,k,niter,stat,newton_steps
  integer :: ndim,nt,n23,nw,nq
  real    :: ow,dw,dt,ot,oky,dky,percentile,t,t2,maxin,quant
  double precision :: res,res0,res1,res2

  double precision :: alpha
  logical          :: icont,cont,giveup,verb

  call sep_init(SOURCE)

  ndim=sep_dimension()
  allocate(n(ndim),o(ndim),d(ndim))

  do i=1,ndim
     call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
  end do

  nt=n(1); dt=d(1); ot=o(1)
  if (ndim.ge.2) then
     n23=product(n(2:))
  else
     n23=1
  end if

  allocate(in(nt,n23),gt(nt,n23))

  call from_param('niter',niter,10)
  call from_param('verb',verb,.false.)
  call from_param('t2',t2,0.)
  call from_param('percentile',percentile,90.)
  if ((percentile.lt.1).or.(percentile.gt.100.)) then
     call erexit('ERROR: percentile must be between 0 and 100')
  end if

  gt=1.
  if (exist_file('gain')) then
     call sreed('gain',gt,4*nt*n23)
  end if

  do i=1,nt
     t=(i-1)*dt+ot
     gt(i,:)=t**t2*gt(i,:)
  end do

  if (exist_file('gt')) then
     call srite('gt',gt,4*nt*n23)
     call to_history('n1',nt,'gt')
     call to_history('d1',dt,'gt')
     call to_history('o1',ot,'gt')
     call to_history('n2',n23,'gt')
     call to_history('d2',d(2),'gt')
     call to_history('o2',o(2),'gt')
  end if

  call sreed('in',in,4*nt*n23)

  nq=size(in)*percentile/100.
  allocate(tmp1(nt*n23))
  do i=1,n23
     do j=1,nt
        tmp1(j+(i-1)*nt)=abs(in(j,i))
     end do
  end do
  quant=quantile_calc(nq,tmp1); deallocate(tmp1)
  write(0,*) 'INFO:',percentile,'th percentile input data=',quant

  nw=nt/2+1

  allocate(plan1(2))
  write(0,*) 'INFO: Initialize FFT plans'
  call init_fft_dimension_1d_r2c_c2r(nt,dt,dw,ow)
  call initialize_fft1d_r2c_c2r(plan1)

  allocate(inw(nw,n23))
  write(0,*) 'INFO: FFT input data'
  call fft1d_r2c_c2r_n23(in,inw,n23,plan1,FORWARD)

  allocate(uw(nw),duw(nw),euw(nw),ut(nt),dut(nt),tmp(nw),tmp2(nw))
  if (exist_file('filter_movie')) allocate(uw1(nw),euw1(nw),ut1(nt))
  allocate(drw(nw),dqt(nt,n23))
  allocate(rt(nt,n23),qt(nt,n23))

  ut=0.
  uw=0.
  rt=in

  call hysdstep_init(nt*n23)

  cont=.true.
  giveup=.false.
  i=0

  allocate(tmp1(nt*n23))
  do i=1,n23
     do j=1,nt
        tmp1(j+(i-1)*nt)=abs(rt(j,i)*gt(j,i))
     end do
  end do
  quant=quantile_calc(nq,tmp1); deallocate(tmp1)
  allocate(tmp1(nt))

  if (quant.eq.0.) then
     write(0,*) 'INFO: quant=',quant
     write(0,*) 'INFO: replace quant with maxval'
     quant=maxval(abs(rt*gt))
  end if
  write(0,*) 'INFO:',percentile,'th percentile gained data=',quant

  gt=gt/quant
  qt=gt*rt                                                   ! q  = g*r

  i=0
  j=0
  do while (cont.and.(.not.giveup))

     call fft1d_r2c_c2r_n23(ut,uw,1,plan1,FORWARD)
     duw=0.
     dut=0.
     dqt=0.
     res=0.

     do j=1,n23

        res = res+sum(sqrt(1.d0+qt(:,j)**2)-1.d0)
        tmp1=gt(:,j)*qt(:,j)/sqrt(qt(:,j)**2+1)              !tmp1= g*softclip(q)
        call fft1d_r2c_c2r_n23(tmp1,tmp,1,plan1,FORWARD)     !tmp = FT(tmp1)
        call fft1d_r2c_c2r_n23(rt(:,j),tmp2,1,plan1,FORWARD) !tmp2= FT(r) 
        duw=duw+conjg(tmp2)*tmp                              !  dU= dU+conjg(FT(r))*FT(g*softclip(q))
     end do
     if(i.eq.0) res0=res
     if(exist_file('fct')) call srite('fct',res*100/res0,4)

     duw=(duw-sum(duw)/nw)                                   ! remove the mean
     call fft1d_r2c_c2r_n23(dut,duw,1,plan1,BACKWARD)        ! du= IFT(dU)

     do j=1,n23
        drw=0.
        call fft1d_r2c_c2r_n23(rt(:,j),tmp2,1,plan1,FORWARD) !tmp2= FT(r)
        drw=tmp2*duw                                         !dR  = FT(r)*dU
        call fft1d_r2c_c2r_n23(tmp1,drw,1,plan1,BACKWARD)    !tmp1= IFT(dR)
        dqt(:,j)=gt(:,j)*tmp1                                ! dq = g*IFT(dR)
     end do

     stat=hysdstep(alpha,qt,dqt,newton_steps)

     if (verb) write(0,*) 'INFO: iter=',i,'/',niter,'Res=',sngl(res*100/res0),'Newton steps',newton_steps

     if (i.ge.niter) cont=.false.
     icont=.true.
     res2=10.
     do while(icont)
        res1=0.
        tmp1=ut+sngl(alpha)*dut
        call fft1d_r2c_c2r_n23(tmp1,uw,1,plan1,FORWARD)
        do j=1,n23

           euw=exp(uw)
           tmp=inw(:,j)*euw
           call fft1d_r2c_c2r_n23(rt(:,j),tmp,1,plan1,BACKWARD) ! r  = IFT(D*exp(U))
           qt(:,j)=gt(:,j)*rt(:,j)                              ! q  = g*r
           res1   = res1+sum(sqrt(1.d0+qt(:,j)**2)-1.d0)
        end do

        if (res1.eq.res2) then
           write(0,*) 'INFO: stuck decreasing alpha, leave iterations now'
           giveup=.true.
           icont=.false.
        else
           res2=res1
           if (res1.lt.res) then
              icont=.false.
           else
              alpha=alpha/2.
           end if
        end if
     end do

     if (.not.giveup) then
        ut=ut+sngl(alpha)*dut
        if (exist_file('movie')) then
           call srite('movie',qt,4*nt*n23)
        end if
        if (exist_file('filter_movie')) then
           call fft1d_r2c_c2r_n23(ut,uw1,1,plan1,FORWARD)           
           euw1=exp(uw1)
           uw1=euw1
           call fft1d_r2c_c2r_n23(ut1,uw1,1,plan1,BACKWARD)
           call srite('filter_movie',ut1,4*nt)
        end if
        if (exist_file('ut_movie')) then
           call srite('ut_movie',cshift(ut,nt/2),4*nt)
        end if
        i=i+1
     end if

  end do

  call srite('out',rt,4*nt*n23)

  if (exist_file('filter')) then
     call fft1d_r2c_c2r_n23(ut,uw,1,plan1,FORWARD)
     euw=exp(uw)
     uw=euw
     call fft1d_r2c_c2r_n23(ut,uw,1,plan1,BACKWARD)
     call srite('filter',ut,4*nt)
     call to_history('n1',nt,'filter')
     call to_history('d1',dt,'filter')
     call to_history('n2',1,'filter')
     call to_history('n3',1,'filter')
  end if

  call destroy_fft(plan1)

  if (exist_file('filter_movie')) then
     call to_history('n1',nt,'filter_movie')
     call to_history('d1',dt,'filter_movie')
     call to_history('o1',ot,'filter_movie')
     call to_history('n2',1,'filter_movie')
     call to_history('n3',i,'filter_movie')
  end if
  if (exist_file('ut_movie')) then
     call to_history('n1',nt,'ut_movie')
     call to_history('d1',dt,'ut_movie')
     call to_history('o1',-dt*(nt-1)/2,'ut_movie')
     call to_history('n2',1,'ut_movie')
     call to_history('n3',i,'ut_movie')
  end if
  if (exist_file('movie')) then
     call to_history('n3',i,'movie')
     call to_history('n2',n23,'movie')
     call to_history('n1',n(1),'movie')
     call to_history('d2',d(2),'movie')
     call to_history('o2',o(2),'movie')
  end if
  if (exist_file('fct')) then
     call to_history('n1',i,'fct')
     call to_history('d1',1,'fct')
     call to_history('o1',1,'fct')
  end if
  call sep_close()

end program Sparse_decon
