program Sparse_decon_reg

  use sep
  use sfft_mod
  use hycdstep_mod
  use quantile_calc_mod

  implicit none

  real,       dimension(:,:),   allocatable :: in, gin ! input data
  real,       dimension(:,:),   allocatable :: gt, wt  ! gain function
  real,       dimension(:,:),   allocatable :: qt,dqt
  real,       dimension(:,:),   allocatable :: rt,rtmp

  real,       dimension(:),     allocatable :: tmp1, regt, quantarray

  complex,    dimension(:,:),   allocatable :: inw, ginw ! input Fourier domain
  complex,    dimension(:,:),   allocatable :: rw
  complex,    dimension(:),     allocatable :: uw, duw, uw1, u0

  double complex, dimension(:), allocatable :: euw, euw1     ! filter and delta filter Fourier domain

  complex,    dimension(:),     allocatable :: tmp,tmp2, regw
  complex,    dimension(:),     allocatable :: drw

  real,       dimension(:),     allocatable :: ut, utb, ut1, dut, delreg ! filter time domain


  integer,    dimension(:),     allocatable :: n
  real,       dimension(:),     allocatable :: o, d
  integer(8), dimension(:),     allocatable :: plan1,plan2
  logical,    parameter                     :: FORWARD=.FALSE.,BACKWARD=.TRUE.

  character(len=1024) :: label

  integer :: it,i,j,k,l,m,niter,stat,newton_steps,lag,nchanges,nalphas,nfreq,nfreq2,count1
  integer :: ndim,nt,n23,nw,nq,nt2
  real    :: ow,dw,dt,ot,oky,dky,percentile,t,t2,maxin,quant,eps,ratio,freg,fdat,weight,qbar
  double precision :: res,res0,res1,res2,resd,resr,resd0,resr0

  double precision :: alpha, alpha_keep
  logical          :: icont,cont,giveup,verb,new_init,zero_neg,reg0

  call sep_init(SOURCE)

  ndim=sep_dimension()
  allocate(n(ndim),o(ndim),d(ndim))

  do i=1,ndim
     call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
  end do

  nt=n(1); dt=d(1); ot=o(1)
  if (ndim.ge.2) then
     n23=product(n(2:))
  else
     n23=1
  end if

  allocate(in(nt,n23),gt(nt,n23),wt(nt,n23),gin(nt,n23))
  call from_param('niter',niter,10)
  call from_param('nfreq',nfreq,10)
  call from_param('nfreq2',nfreq2,5)
  call from_param('lag',lag,12)
  call from_param('ratio',ratio,100.)
  call from_param('nchanges',nchanges,0)
  call from_param('nalphas',nalphas,niter)
  call from_param('verb',verb,.false.)
  call from_param('reg0',reg0,.false.)
  call from_param('zero_neg',zero_neg,.false.)
  call from_param('new_init',new_init,.false.)
  call from_param('t2',t2,0.)
  call from_param('percentile',percentile,90.)
  if ((percentile.lt.1).or.(percentile.gt.100.)) then
     call erexit('ERROR: percentile must be between 0 and 100')
  end if

  gt=1.
  if (exist_file('gain')) then
     call sreed('gain',gt,4*nt*n23)
  end if

  nw=nt/2+1
  nt2=nt/2

  if (exist_file('u0')) then
     write(0,*) 'INFO: Found starting guess u0'
     allocate(u0(nw))
     call sreed('u0',u0,8*nw)
  end if

  do i=1,nt
     t=(i-1)*dt+ot
     wt(i,:)=t**t2
  end do

  if (exist_file('gt')) then
     call srite('gt',gt*wt,4*nt*n23)
     call to_history('n1',nt,'gt')
     call to_history('d1',dt,'gt')
     call to_history('o1',ot,'gt')
     call to_history('n2',n23,'gt')
     call to_history('d2',d(2),'gt')
     call to_history('o2',o(2),'gt')
  end if

  call sreed('in',in,4*nt*n23)

  gin=wt*in

  nq=size(in)*percentile*0.01
  allocate(quantarray(nt*n23))
  do i=1,n23
     do j=1,nt
        quantarray(j+(i-1)*nt)=abs(in(j,i))
     end do
  end do
  quant=quantile_calc(nq,quantarray)
  write(0,*) 'INFO:',percentile,'th percentile input data=',quant

  allocate(plan1(2),plan2(2))
  write(0,*) 'INFO: Initialize FFT plans'
  call init_fft_dimension_1d_r2c_c2r(nt,dt,dw,ow)
  call initialize_fft1d_r2c_c2r(plan1)
  call initialize_fft1d_double_r2c_c2r(plan2)

  allocate(inw(nw,n23))
  allocate(ginw(nw,n23))
  write(0,*) 'INFO: FFT input data'
  call fft1d_r2c_c2r_n23(in,inw,n23,plan1,FORWARD)
  call fft1d_r2c_c2r_n23(gin,ginw,n23,plan1,FORWARD)

  allocate(uw(nw),duw(nw),euw(nw),ut(nt),utb(nt),dut(nt),regt(nt),delreg(nt),regw(nw),tmp(nw),tmp2(nw))
  if (exist_file('filter_movie')) allocate(uw1(nw),euw1(nw),ut1(nt))
  allocate(drw(nw),dqt(nt,n23))
  allocate(rt(nt,n23),qt(nt,n23),rtmp(nt,n23))

  ut=0.
  utb=0.
  uw=0.
  tmp=0.

  ! Forming the regularization array utb
  do j=1,n23
     tmp=tmp+ginw(:,j)*conjg(ginw(:,j))
  end do
  tmp=sqrt(real(tmp))
  uw=clog(tmp)
  uw=uw-sum(uw)/nw
  call fft1d_r2c_c2r_n23(utb,uw,1,plan1,BACKWARD) 
  call mostly_causal_amplitude(utb,nt,lag)  
  call fft1d_r2c_c2r_n23(utb,uw,1,plan1,FORWARD)
  deallocate(gin,ginw)
  uw=-uw
  call fft1d_r2c_c2r_n23(utb,uw,1,plan1,BACKWARD)

  if (exist_file('utb')) then
     call srite('utb',uw,8*nw)
     call to_history('n1',nw,'utb')
     call to_history('d1',dw,'utb')
     call to_history('o1',ot,'utb')
     call to_history('n2',1,'utb')
     call to_history('n3',1,'utb')
     call to_history('nt',nt,'utb')
     call to_history('ot',ot,'utb')
     call to_history('dt',dt,'utb')
     call to_history('esize',8,'utb')
  end if

  if (new_init) then
     ut=utb
  else
     uw=0.0 
  end if

  do j=1,n23
     tmp=inw(:,j)*exp(uw)
     call fft1d_r2c_c2r_n23(rt(:,j),tmp,1,plan1,BACKWARD) ! r  = IFT(D*exp(U))
  end do

  call hycdstep_init(nt*n23)
  call hycdstep_nreg_init(nt)

  cont=.true.
  giveup=.false.
  i=0

  allocate(tmp1(nt*n23))
  do i=1,n23
     do j=1,nt
        tmp1(j+(i-1)*nt)=abs(rt(j,i)*wt(j,i))
     end do
  end do
  quant=quantile_calc(nq,tmp1); deallocate(tmp1)
  allocate(tmp1(nt))

  if (quant.eq.0.) then
     write(0,*) 'INFO: quant=',quant
     write(0,*) 'INFO: replace quant with maxval'
     quant=maxval(abs(rt*wt))
  end if
  write(0,*) 'INFO:',percentile,'th percentile gained data=',quant

  if (reg0) then
     write(0,*) 'INFO: Regularization eps(u(t)-\bar(u(t)))^2'
  else
     write(0,*) 'INFO: Regularization eps(u(t)-u(-t))^2'
  end if

  if (zero_neg) write(0,*) 'INFO: negative lags are zeroed'

  if (allocated(u0)) then
     euw=exp(u0)
     do j=1,n23
        tmp=inw(:,j)*euw
        call fft1d_r2c_c2r_n23(rt(:,j),tmp,1,plan1,BACKWARD) ! r  = IFT(D*exp(U))
     end do
     uw=u0
     deallocate(u0)
     call srite('rtinit',rt,4*nt*n23)
     call to_history('n2',n23,'rtinit')
     call to_history('n1',n(1),'rtinit')
     call to_history('d2',d(2),'rtinit')
     call to_history('o2',o(2),'rtinit')
  end if

  call from_param('quant',quant,quant)
  write(0,*) 'INFO: quantile read from param=',quant

  wt=wt/quant
  qt=wt*gt*rt                                                ! q  = g*r

  i=0
  j=0
  regt=0.

  do j=1,n23
     fdat = sum(sqrt(1.d0+qt(:,j)**2)-1.d0)
  end do
  if (reg0) then
     regt = ut-utb
  else
     do it=2,nt2
        regt(it)        =ut(it)-ut(2*nt2-it+2)
        regt(2*nt2-it+2)=ut(2*nt2-it+2)-ut(it)
     end do
     regt(1)=0.    ! Zero freq
     regt(nt2+1)=0.! Nyquist
  end if
  call from_param('eps',eps,1.)     
  write(0,*) 'INFO: freg=0, eps read from param file=',eps

  if (exist_file('movie')) then
     call srite('movie',rt,4*nt*n23)
  end if
  if (exist_file('filter_movie')) then         
     call srite('filter_movie',uw,8*nw)
  end if

  count1=0

  do while (cont.and.(.not.giveup))

     duw=0.
     dut=0.
     dqt=0.
     res=0.
     resr=0.
     resd=0.
     regt=0.
     ut=0.

     call fft1d_r2c_c2r_n23(ut,uw,1,plan1,BACKWARD)
     if (reg0) then
        regt=ut-utb
     else
        do it=2,nt2
           regt(it)        =ut(it)-ut(2*nt2-it+2)
           regt(2*nt2-it+2)=ut(2*nt2-it+2)-ut(it)
        end do
        regt(1)=0.     ! Zero freq
        regt(nt2+1)=0. ! Nyquist
     end if
     
     resr=eps*sum(dprod(regt,regt))*0.5
     if (i.eq.0) resr0=resr
     res=res+resr

     do j=1,n23
        resd = resd+sum(sqrt(1.d0+qt(:,j)**2)-1.d0)
        tmp1=wt(:,j)*gt(:,j)*qt(:,j)/sqrt(qt(:,j)**2+1)      !tmp1= g*softclip(q)
        call fft1d_r2c_c2r_n23(tmp1,tmp,1,plan1,FORWARD)     !tmp = FT(tmp1)
        call fft1d_r2c_c2r_n23(rt(:,j),tmp2,1,plan1,FORWARD) !tmp2= FT(r) 
        duw=duw+conjg(tmp2)*tmp                              !  dU= dU+conjg(FT(r))*FT(g*softclip(q))
     end do

     if(i.eq.0) resd0=resd
     res=resd+resr
     if(i.eq.0) res0=res
     if(exist_file('fct')) call srite('fct',sngl(res),4)
     if(exist_file('qbar')) then 
        qbar=0
        do j=1,n23
           qbar=qbar+sum(sqrt(1.+qt(:,j)**2))
        end do
        qbar=sqrt((qbar/(nt*n23))**2-1)
        call srite('qbar',qbar,4)
     end if

     call fft1d_r2c_c2r_n23(dut,duw,1,plan1,BACKWARD)  
     dut=dut+eps*regt
     !call srite('dut',eps*regt,4*nt)
     
     if (zero_neg) then !dut(nt/2+1:nt-lag)=0.
        do it=2,lag
           weight=cos(.5*3.14159*(it-1)/(lag-1.))**2
!           weight=1.
           dut(nt-it+2) = weight*dut(nt-it+2) 
        end do
        dut(nt/2+2:nt-lag+1)=0.
     end if
     call fft1d_r2c_c2r_n23(dut,duw,1,plan1,FORWARD)
     duw=(duw-sum(duw)/nw)   
     call fft1d_r2c_c2r_n23(dut,duw,1,plan1,BACKWARD)     
     
     do j=1,n23
        drw=0.
        call fft1d_r2c_c2r_n23(rt(:,j),tmp2,1,plan1,FORWARD) !tmp2= FT(r)
        drw=tmp2*duw                                         !dR  = FT(r)*dU
        call fft1d_r2c_c2r_n23(tmp1,drw,1,plan1,BACKWARD)    !tmp1= IFT(dR)
        dqt(:,j)=wt(:,j)*gt(:,j)*tmp1                        ! dq = g*IFT(dR)
     end do

     if (reg0) then
        delreg=dut
     else
        do it=2,nt2
           delreg(it)        =dut(it)-dut(2*nt2-it+2)
           delreg(2*nt2-it+2)=dut(2*nt2-it+2)-dut(it)
        end do
        delreg(1)=0.      ! Zero freq
        delreg(nt2+1)=0.  ! Nyquist
     end if

     if (i.lt.nalphas) stat=hycdstep_reg(alpha,qt,dqt,regt,delreg,eps,newton_steps)
 
     if (verb) then
        if (mod(i,nfreq).eq.0) then
           write(0,*) 'INFO: # iter=',i,'/',niter,'Newton steps',newton_steps
           write(0,*) '      # Res=',sngl(resd+resr),'Res d=',sngl(resd),'Res reg',sngl(resr)
           write(0,*) '      # Alp=',alpha
        end if
     end if

     if (i.ge.niter) cont=.false.
     icont=.true.
     res2=res     
     do while(icont)
        resd=0.
        resr=0.
        regt=0.  
        res1=0.
        tmp1=0.
        tmp2=0.
        tmp=0.
        tmp=uw+sngl(alpha)*duw
        euw=exp(tmp)       
        call fft1d_r2c_c2r_n23(tmp1,tmp,1,plan1,BACKWARD)
        if (reg0) then
           regt=tmp1-utb
        else
           do it=2,nt2
              regt(it)        =tmp1(it)-tmp1(2*nt2-it+2)
              regt(2*nt2-it+2)=tmp1(2*nt2-it+2)-tmp1(it)
           end do      
           regt(1)=0.     ! Zero freq
           regt(nt2+1)=0. ! Nyquist     
        end if
        resr=eps*sum(dprod(regt,regt))*0.5
        do j=1,n23
           tmp2=inw(:,j)*euw
           call fft1d_r2c_c2r_n23(rt(:,j),tmp2,1,plan1,BACKWARD)! r  = IFT(D*exp(U))
           qt(:,j)=wt(:,j)*gt(:,j)*rt(:,j)                              ! q  = g*r
           resd   = resd+sum(sqrt(1.d0+qt(:,j)**2)-1.d0)
        end do
        res1=resr+resd

        if (i.lt.nalphas) then
           call fft1d_r2c_c2r_n23(dut,duw,1,plan1,BACKWARD)
           res=res+0.0001*alpha*sum(dprod(dut,dut)) ! Armijo rule
           if (res1.eq.res2) then
              write(0,*) 'INFO: stuck decreasing alpha, leave iterations now'
              giveup=.true.
              icont=.false.
           else
              res2=res1
              if (res1.le.res) then
                 icont=.false.
              else
                 alpha=alpha*0.5
              end if
           end if
        else 
           icont=.false.
           if (res1.eq.res2) then
              write(0,*) 'INFO: stuck decreasing alpha, leave iterations now'
              giveup=.true.
           else
              alpha=alpha_keep
           end if
        end if
     end do

     alpha_keep=alpha

     if (.not.giveup) then
        uw=uw+sngl(alpha)*duw
        if (exist_file('movie')) then
           call srite('movie',rt,4*nt*n23)
        end if
        if (exist_file('qt_movie')) then
           rtmp=qt
           do j=1,50
              rtmp(1:50,j)=qbar
           end do
           do j=n23-50,n23
              rtmp(1:50,j)=-qbar
           end do
           call srite('qt_movie',rtmp,4*nt*n23)
        end if
        if (exist_file('filter_movie')) then         
           call srite('filter_movie',uw,8*nw)
        end if
        i=i+1
     end if

     if ((mod(i,nfreq2).eq.0).and.(count1.lt.nchanges)) then
        count1=count1+1
        wt=wt*quant
        do k=1,n23
           do l=1,nt
              quantarray(l+(k-1)*nt)=abs(wt(l,k)*rt(l,k))
           end do
        end do
        quant=quantile_calc(nq,quantarray)
        write(0,*) 'INFO:',percentile,'th percentile residual',quant
        wt=wt/quant
        qt=wt*gt*rt
     end if

  end do

  write(0,*) 'INFO: Inversion finished after',i,'iterations'
  call srite('out',rt,4*nt*n23)

  if (exist_file('filter')) then
     call srite('filter',uw,8*nw)
     call to_history('n1',nw,'filter')
     call to_history('d1',dw,'filter')
     call to_history('n2',1,'filter')
     call to_history('n3',1,'filter')
     call to_history('nt',nt,'filter')
     call to_history('ot',ot,'filter')
     call to_history('dt',dt,'filter')
     call to_history('esize',8,'filter')
  end if
  if (exist_file('filter_movie')) then
     call to_history('n1',nw,'filter_movie')
     call to_history('d1',dw,'filter_movie')
     call to_history('o1',ot,'filter_movie')
     call to_history('n2',1,'filter_movie')
     call to_history('n3',i+1,'filter_movie')
     call to_history('nt',nt,'filter_movie')
     call to_history('ot',ot,'filter_movie')
     call to_history('dt',dt,'filter_movie')
     call to_history('esize',8,'filter_movie')

  end if
  if (exist_file('movie')) then
     call to_history('n3',i+1,'movie')
     call to_history('n2',n23,'movie')
     call to_history('n1',n(1),'movie')
     call to_history('d2',d(2),'movie')
     call to_history('o2',o(2),'movie')
  end if
  if (exist_file('qt_movie')) then
     call to_history('n3',i,'qt_movie')
     call to_history('n2',n23,'qt_movie')
     call to_history('n1',n(1),'qt_movie')
     call to_history('d2',d(2),'qt_movie')
     call to_history('o2',o(2),'qt_movie')
  end if
  if (exist_file('fct')) then
     call to_history('n1',i,'fct')
     call to_history('d1',1,'fct')
     call to_history('o1',1,'fct')
  end if
  if (exist_file('qbar')) then
     call to_history('n1',i,'qbar')
     call to_history('d1',1,'qbar')
     call to_history('o1',1,'qbar')
  end if
!  call to_history('n1',nt,'dut')
!  call to_history('o1',ot,'dut')
!  call to_history('d1',dt,'dut')
!  call to_history('n2',1,'dut')
!  call to_history('n3',i,'dut')
  call sep_close()

end program Sparse_decon_reg

subroutine mostly_causal_energy(ut,nt)
  integer            ::     nt
  real, dimension(nt)::  ut

  integer :: it,lag, nt2
  real    :: from, weight
  
  nt2=nt/2
  ut(1    )  = ut(1    ) / 2.          
  ut(1+nt2)  = ut(1+nt2) / 2. 
  ut(2+nt2:) = 0.

  ! BEGIN stuff added to remove a little of the asymmetric part.
  lag = 15
  do it = 2, lag 
     from   = ut(it)
     weight = cos( .5* 3.1416 * (it-1.)/(lag-1.))**2
     ut(it)        =  ut(it)      - weight * from /2.
     ut(2*nt-it+2) =  ut(2*nt-it+2) + weight * from /2.
  end do

end subroutine mostly_causal_energy

subroutine mostly_causal_amplitude(ut,nt,lag)
  integer            ::               nt,lag
  real, dimension(nt)::            ut

  integer :: it,nt2
  real    :: weight, asym

  nt2=nt/2
  do it=2,nt2
     ut(it)         = 2.*ut(it)
     ut(2*nt2-it+2) = 0.
  end do

  do it=2,lag
     asym=(ut(it)-ut(2*nt2-it+2))/2.
     weight=cos(.5*3.14159*(it-1)/(lag-1.))**2
!     weight=1
     ut(      it)   = ut(      it)   - weight*asym
     ut(2*nt2-it+2) = ut(2*nt2-it+2) + weight*asym
  end do
   
end subroutine mostly_causal_amplitude
