program Wavelet_Estimation

  use omp_lib

  use sep
  use mute
  use helix
  use envelope
  use irls_priv
  use verbosity_mod
  use adaptive_init
  use findadaptive_mod

  implicit none

  integer, external                    :: getch
  real,    dimension(:,:), allocatable :: synt_data,obs_data,weight_obs_data,weight_synth_data,est_wavelet
  real,    dimension(:), allocatable   :: input_wavelet, output_wavelet
  real,    dimension(:,:), allocatable :: output_wavelet_individual
  real,    dimension(:), allocatable   :: mask
  type (filter)                        :: match_n,laplac_n, match_tmp

  integer :: wmode                       ! 1=noweight 2=L1
  integer :: ndim                        ! dimension of dataset
  integer :: ndim_domain                 ! dimension for the filtering (will be 2 here)
  integer :: nfreq                       ! set >0 for l1 via nonlinear CG
  integer :: nmem, niter
  real    :: rbar                        ! rbar for l1
  real    :: maxval1, maxval2, maxval3   ! For pre-normalization

  character (len=20) :: targetname
  integer, dimension(:), pointer :: n                  ! data-size, dimension(ndim)
  real,    dimension(:), pointer :: d, o

  integer, dimension(:), pointer :: idx                ! index counter (ndim)

  integer :: i,stat,ilag,ipatch,idim,ipanels,npanels   ! junk variables
  integer :: n12, n1w
  logical :: weight_synt_data,mute_me
  integer :: ncoeff_n, tc1 ,tc2, verb
  real    :: vmute, tmute
  logical :: want_filter ! If true, outputs the filter

  integer :: sided ! 1 means one-sided filter with
                   ! leading coef on the left
                   ! 2 means two-sided filter with 
                   ! leading coef in the center

  integer :: num_cpu, wait_rd, wait_wt, wait_for_read_thread, wait_for_rite_thread
  integer :: ithread, my_pid, count
  character(len=1024)    :: label

  integer, dimension(:), allocatable :: status,nfilt_n
  real, dimension(:), allocatable    :: scale_dat, scale_nois

  call sep_init(SOURCE)
 
  write(0,*) 'INFO:'
  write(0,*) 'INFO:'
  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO: Wavelet estimation with time-domain '
  write(0,*) 'INFO:           matching filters          '
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  ndim=sep_dimension()
  ndim_domain=2      ! Filters work on a 2D plane only
  allocate(nfilt_n(1))

  if (ndim.le.2) then
     allocate(n(3), o(3), d(3))
     do i=1,ndim
        call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
     end do
     n(ndim+1:)=1
     o(ndim+1:)=0.
     d(ndim+1:)=1
  else
     allocate(n(ndim), o(ndim), d(ndim))
     do i=1,ndim
        call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
     end do
  end if

  if (exist_file('input_wavelet')) then
     call from_aux('input_wavelet','n1',n1w)
     allocate(input_wavelet(n1w))
     allocate(output_wavelet(n1w)); output_wavelet=0.
     call sreed('input_wavelet',input_wavelet,4*n1w)
     write(0,*) 'INFO: Done reading input_wavelet'
  else
     call erexit('ERROR: need input_wavelet file')
  end if

  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:       Dimensions Observed Data      '
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  do i=1,ndim
     write(0,*) 'INFO: - - - - - - - '
     write(0,*) 'INFO: n(',i,')=',n(i)
     write(0,*) 'INFO: o(',i,')=',o(i)
     write(0,*) 'INFO: d(',i,')=',d(i)
  end do

  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  npanels=product(n(3:))
  n12=n(1)*n(2)
  count=npanels

  call from_param('num_cpu',num_cpu,omp_get_max_threads())
  call from_param('want_filter',want_filter,.false.)
  
  if(want_filter) then
    if(.not.exist_file('filter')) then
      call erexit('ERROR: Need filter file')
    end if
  end if

  if (npanels.lt.num_cpu) then
     num_cpu=min(npanels,num_cpu)
     write(0,*) 'WARN: # of panels< # of threads'
     write(0,*) 'WARN: num_cpu changed to # of panels'
  end if

  if (num_cpu.gt.omp_get_max_threads()) then
     num_cpu=omp_get_max_threads()
     write(0,*) 'WARN: num_cpu > max # threads'
     write(0,*) 'WARN: change num_cpu to max '
  end if
  
  allocate(status(num_cpu),scale_dat(num_cpu),scale_nois(num_cpu))

  call from_param('wmode',wmode,1)    ! 1=no weight, 2=L1 norm with Iteratively Reweighted Least Squares (IRLS)
  call from_param('verb',verb,1)

  if(verb.gt.2) then
     if (num_cpu.gt.1) then
        verb=2
        write(0,*) 'WARN: num_cpu and verb are incompatible'
        write(0,*) 'WARN: change verb to 2'
     end if
  end if

  ! Read synt_data filter paramaters
  call from_param('nfilt',nfilt_n,(/21/))
  call from_param('niter',niter,product(nfilt_n)-1)
  call from_param('sided',sided,1) ! defaults to one-sided filter

  call from_param('weight_synt_data',weight_synt_data,.false.)

  ! Read Mute parameters
  call from_param('mute',mute_me,.false.)
  if (mute_me) then
     call from_param('vmute',vmute,1500.)
     call from_param('tmute',tmute,o(1))
     call mute_init(vmute,tmute,n(2),d(2),o(2),n(1),d(1),o(1))
  end if

  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:        Inversion parameters         '
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO: Number of iterations        =',niter
  write(0,*) 'INFO: Size of filter in # samples =',nfilt_n(1)
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  if (mute_me) then
     if (verb>1) write(0,*) 'INFO:'
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:         Muting parameters           '
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO: Velocity mute function =',vmute
     if (verb>1) write(0,*) 'INFO: T0 of the mute function=',tmute
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:'
  end if

  allocate(synt_data(n12,num_cpu),obs_data(n12,num_cpu),est_wavelet(n1w,num_cpu))

  if (mute_me) then
     ! build a simple masking operator from vmute and tmute
     ! this mask is the same for all gathers.
     allocate(mask(n12))
     call mute_build(mask)
     if (exist_file('mute_out')) then
        call to_history ("n1", n(1),   "mute_out")
        call to_history ("n2", n(2),   "mute_out")
        call to_history ("n3", 1,   "mute_out")
        call srite("mute_out",mask,n12*4)
     end if
  end if

  if (wmode.eq.3.or.weight_synt_data) then
     ! Initialize the envelope weighting function for 
     ! the data residual
     call from_param('tc1',tc1,10)
     call from_param('tc2',tc2,10)
     if ((tc1.gt.n(1)).or.(tc2.gt.n(2))) call seperr('ERRO: tc1 or tc2 too big')
     call envelope_init(n(1:2),ndim_domain,tc1,tc2)
     if(wmode.eq.3) allocate(weight_obs_data(n12,num_cpu)) 
     if(exist_file("output_weight_synt_data")) allocate(weight_synth_data(n12,num_cpu))
     if (verb>1) write(0,*) 'INFO:'
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:          Weighting parameters       '
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO: wmode                              =',wmode
     if (verb>1) write(0,*) 'INFO: weight_synt_data                   =',weight_synt_data
     if (verb>1) write(0,*) 'INFO: Size smoothing axis 1 (in samples) =',tc1
     if (verb>1) write(0,*) 'INFO: Size smoothing axis 2 (in samples) =',tc2
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:'
  end if

  if (wmode.eq.2) then
     call from_param('rbar',rbar,0.01) 
     stat=l1_priv_init(rbar)
     call from_param('nfreq',nfreq,5)
     call from_param('nmem',nmem,0)

     if ((nmem.ge.nfreq).or.(nfreq.gt.niter)) call seperr('ERRO: nfreq or nmem have wrong values')
     if (rbar.gt.1) call seperr('ERRO: rbar should be less or equal to 1')

     if (verb>1) write(0,*) 'INFO:'
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:         IRLS L1 parameters          ' 
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO: # iters between restarts =',nfreq
     if (verb>1) write(0,*) 'INFO: # iters before using L1  =',nmem
     if (verb>1) write(0,*) 'INFO: L1/L2 threshold value    =',rbar
     if (verb>1) write(0,*) 'INFO: ------------------------------------'
     if (verb>1) write(0,*) 'INFO:'

  end if

  call create_match_filter_sided(nfilt_n,match_tmp,1,n,sided)
     
  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO: Starting processing of ',npanels,' panels using',num_cpu,' threads'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO:'

  wait_for_read_thread=1
  wait_for_rite_thread=1
  call omp_set_num_threads(num_cpu)

  !$omp  parallel do &
  !$omp& schedule( static, 1 ) &
  !$omp& default( shared ) &
  !$omp& private( ithread, ipanels, match_n, ncoeff_n, my_pid, wait_rd, wait_wt) &
  !$omp& firstprivate( npanels, num_cpu )

  thread_loop:do ithread=1,num_cpu

     my_pid=omp_get_thread_num()

     status(ithread)=1
     scale_dat=1;scale_nois=1
     obs_data(:,ithread)=0;synt_data(:,ithread)=0;est_wavelet(:,ithread)=0.
     if(wmode.eq.3) weight_obs_data(:,ithread)=0
     if(exist_file("output_weight_synt_data")) weight_synth_data(:,ithread)=0

     panels_loop:do ipanels=ithread,npanels,num_cpu

        ! READ PART
        wait_rd=1
        do while(wait_rd.eq.1)
           
           !$omp critical
           if (ipanels.eq.wait_for_read_thread) then
              call sreed("in",synt_data(:,ithread),n12*4)
              call sreed("observed_data",obs_data(:,ithread),n12*4)
              
              !$omp atomic
              wait_for_read_thread=wait_for_read_thread+1
              
              wait_rd=0
              
           end if
           !$omp end critical

        end do

        ! PROCESSING PART
        call create_copy_match_filter(nfilt_n,match_tmp,match_n)
        call check_array(synt_data(:,ithread),status(ithread))
        call check_array(obs_data(:,ithread),status(ithread))

        STATUS_GATHER:if (status(ithread).eq.1) then

           if (weight_synt_data) then   
              if (exist_file("output_weight_synt_data")) then 
                 call apply_envelope_scaling(synt_data(:,ithread),weight_synth_data(:,ithread))
              else 
                 call apply_envelope_scaling(synt_data(:,ithread))
              end if
              if (mute_me) synt_data(:,ithread)=mask*synt_data(:,ithread)            
           end if

           if (wmode.eq.3) then
              weight_obs_data(:,ithread)=0.
              call comp_envelope(obs_data(:,ithread),weight_obs_data(:,ithread))
              where (weight_obs_data(:,ithread).ne.0)
                 weight_obs_data(:,ithread)=1/(weight_obs_data(:,ithread)+maxval(weight_obs_data(:,ithread))/100)
              end where
           end if

           est_wavelet(:,ithread)=input_wavelet
                         
           select case(wmode)
            
           case(1) ! No weight

              if (verb>2) write(0,*) 'INFO:  -----------------------------------'
              if (verb>2) write(0,*) 'INFO:  Calling findmatch with no weight...'
              if (verb>2) write(0,*) 'INFO:  -----------------------------------'
              
              if (mute_me) then
                 call findmatch(obs_data(:,ithread),synt_data(:,ithread),est_wavelet(:,ithread),match_n,niter,wt=mask,verb=verb)
              else
                 call findmatch(obs_data(:,ithread),synt_data(:,ithread),est_wavelet(:,ithread),match_n,niter,verb=verb)
              end if
              
           case(2) ! IRLS L1
              
              if (verb>2) write(0,*) 'INFO:  ----------------------------'
              if (verb>2) write(0,*) 'INFO:  Calling findmatch with L1...'
              if (verb>2) write(0,*) 'INFO:  ----------------------------'
              
              if (mute_me) then
                 call findmatch(obs_data(:,ithread),synt_data(:,ithread),est_wavelet(:,ithread),match_n,niter,nmem_in=nmem, &
                 & nfreq_in=nfreq,wt=mask,verb=verb)
              else
                 call findmatch(obs_data(:,ithread),synt_data(:,ithread),est_wavelet(:,ithread),match_n,niter,nmem_in=nmem, &
                 & nfreq_in=nfreq,verb=verb)
              end if
              
           case(3) ! Envelope weight(:,ithread)
              
              if (verb>2) write(0,*) 'INFO:  --------------------------------------'
              if (verb>2) write(0,*) 'INFO:  Calling findmatch with envelope weigth'
              if (verb>2) write(0,*) 'INFO:  --------------------------------------'
              
              if (mute_me) weight_obs_data(:,ithread)=weight_obs_data(:,ithread)*mask
              
              call findmatch(obs_data(:,ithread),synt_data(:,ithread),est_wavelet(:,ithread),match_n,niter,wt=weight_obs_data(:,ithread),verb=verb)
              
           case default
              call seperr('ERRO: Weight mode not implemented')
           end select
         
        else
           
           ! DON'T DO ANYTHING IF THERE IS A PROBLEM
           !$omp critical
           count=count-1
           !$omp end critical
           
        end if STATUS_GATHER

        ! WRITE PART
        wait_wt=1
        do while(wait_wt.eq.1)
           
           !$omp critical
           if (ipanels.eq.wait_for_rite_thread) then
              if (wmode.eq.3) then
                 if (exist_file("output_weight_obs_data")) then
                    if (mute_me) weight_obs_data(:,ithread)=weight_obs_data(:,ithread)*mask
                    call srite("output_weight_obs_data",weight_obs_data(:,ithread),n12*4)
                 end if
              end if
              if (weight_synt_data) then
                 if (exist_file("output_weight_synt_data")) then
                    if (mute_me) weight_synth_data(:,ithread)=weight_synth_data(:,ithread)*mask
                    call srite("output_weight_synt_data",weight_synth_data(:,ithread),n12*4)
                 end if
              end if
              if (want_filter) then
                 call srite('filter',match_n%flt,nfilt_n(1)*4)
              end if
              write(0,*) 'INFO: PROGRESS:',real(ipanels*100)/npanels,'%'
              
              !$omp atomic
              wait_for_rite_thread=wait_for_rite_thread+1
              
              wait_wt=0
              
           end if

           !$omp end critical

        end do

        !$omp critical
        output_wavelet=output_wavelet+est_wavelet(:,ithread)
        !$omp end critical

     end do panels_loop
     
  end do thread_loop
  !$omp end parallel do

  call srite("out",output_wavelet/count,n1w*4)
 
  write(0,*) 'INFO:'
  write(0,*) 'INFO: ------------------------------------'
  write(0,*) 'INFO: Done with processing'
  write(0,*) 'INFO: ------------------------------------'

  call to_history('n1',n1w,'out')
  call to_history('d1',d(1),'out')
  call to_history('o1',o(1),'out')
  call to_history('n2',1,'out')
  call to_history('d2',d(2),'out')
  call to_history('o2',o(2),'out')
  call to_history('n3',1,'out')
  call to_history('o3',o(3),'out')
  call to_history('d3',d(3),'out')

  if(want_filter) then
     call to_history('n1',nfilt_n,'filter')
     call to_history('d1',d(1),'filter')
     call to_history('n2',npanels,'filter')
  end if

  if (wmode.eq.3) then
     if (exist_file('output_weight_obs_data')) then
        do i=1,ndim
           call sep_put_data_axis_par("output_weight_obs_data",i,n(i),o(i),d(i),label)
        end do
     end if
  end if

  if(weight_synt_data) then
     if (exist_file('output_weight_synt_data')) then
        do i=1,ndim
           call sep_put_data_axis_par("output_weight_synt_data",i,n(i),o(i),d(i),label)
        end do
     end if
  end if

call sep_close()

end program WAVELET_ESTIMATION
