module findadaptive_mod
  
!  use omp_lib

  use helix
  use helicon_omp
  use nchconest
  use nchelicon
  use cgstep_omp_mod
  use solver2_mod
  use verbosity_mod
  
  ! Weight = L1
  use irls
  use irls_priv

  implicit none

contains


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Find a single matched-filter     !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Minimizes residual:
  ! wght ( synt_data * match - obs_data ) = 0
  subroutine findmatch(obs_data,synt_data,est_wavelet,match,niter,nmem_in,nfreq_in,wt,verb)
    optional                              :: nmem_in,nfreq_in
    real, dimension(:)                    :: obs_data,synt_data,est_wavelet
    real, dimension(:),optional           :: wt
    integer                               :: niter,nmem_in,nfreq_in
    type (filter)                         :: match
    integer                               :: i,nmem,nfreq,type
    logical                               :: l1=.false.,we=.false.,wel1=.false.
    integer                               :: verb
    logical                               :: sverb

    real, dimension(:), allocatable       :: tmp_wavelet
    
    allocate(tmp_wavelet(size(est_wavelet)))
    sverb=.false.
    sverb=(verb>2)

    if (present(nfreq_in).and.present(wt)) then 
       nmem=nmem_in
       nfreq=nfreq_in
       wel1=.true.
    else if (present(wt)) then
       we=.true.
    else if (present(nfreq_in)) then
       nmem=nmem_in
       nfreq=nfreq_in
       l1=.true.
    end if

    if (l1) then
       type=3
       call cgfindmatch(synt_data,obs_data,match,niter,type,sverb,nmem=nmem,nfreq=nfreq)
    else if (we) then
       type=2
       call cgfindmatch(synt_data,obs_data,match,niter,type,sverb,wt=wt)
    else if (wel1) then
       type=4
       call cgfindmatch(synt_data,obs_data,match,niter,type,sverb,wt=wt,nmem=nmem,nfreq=nfreq)
    else
       type=1
       call cgfindmatch(synt_data,obs_data,match,niter,type,sverb)
    end if
    if (sverb) call dumpfilter(match)
    tmp_wavelet=est_wavelet
    est_wavelet=0.
    call applymatch(tmp_wavelet,est_wavelet,match) ! not very elegant, I should have a tmp array
    deallocate(tmp_wavelet)
   
  end subroutine findmatch

  subroutine cgfindmatch(synt_data,obs_data,match,niter,type,sverb,nmem,nfreq,wt)
    optional                              :: nmem,nfreq
    real, dimension(:)                    :: synt_data,obs_data
    real, dimension(:),optional           :: wt
    integer                               :: niter,nmem,nfreq,type
    type (filter)                         :: match
    integer                               :: stat
    logical                               :: sverb

    call nchconest_init(synt_data,match)
    
    select case(type)
    case (1) 
       call solver(nchconest_lop,cgstep_omp,match%flt,obs_data,niter,verb=sverb)
    case (2) 
       call solver(nchconest_lop,cgstep_omp,match%flt,obs_data,niter,verb=sverb,wt=wt)
    case (3) 
       call solver(nchconest_lop,cgstep_omp,match%flt,obs_data,niter,verb=sverb,wght=l1_priv, &
            &     nmem=nmem,nfreq=nfreq)
    case (4) 
       call solver(nchconest_lop,cgstep_omp,match%flt,obs_data,niter,verb=sverb,wt=wt,wght=l1_priv, &
            &     nmem=nmem,nfreq=nfreq)
    end select
   
    call cgstep_omp_close()
    call nchconest_close()

  end subroutine cgfindmatch

  !! Calculates:
  !! output = input * match
  subroutine applymatch(input,output,match)
    
    real, dimension(:)          :: input, output
    type (filter)               :: match
    integer stat
    
    call nchelicon_init(match)
    stat=nchelicon_lop(.false.,.false.,input,output)
 
  end subroutine applymatch

  
  
end module findadaptive_mod












